#ifndef ACCU_HH
# define ACCU_HH

#include <mln/accu/internal/base.hh>

#include "irange.hh"

namespace mln
{
    namespace accu
    {
        template <typename T>
            struct imed : public mln::accu::internal::base< unsigned, imed<T> >
        {
            typedef T argument;

            imed() {}

            void init()
            {
                h_.resize(mln_max(T) + 1, 0);
                nitems_ = 0;
            }

            void take(const argument& v)
            {
                h_[v] += 1;
                nitems_ += 1;
            }

            void take(const imed& other) {}

            unsigned to_result() const
            {
                bool forget_it_;
                return compute_now(forget_it_);
            }

            unsigned compute_now(bool& plus_half) const
            {
                plus_half = false;
                unsigned i = 0, n = mln_max(T) + 1;
                unsigned sum = h_[0], mid = (nitems_ + 1) / 2;
                while (i < n && sum < mid)
                {
                    ++i;
                    sum += h_[i];
                }
                if (nitems_ % 2) // nitems_ is odd.
                    return i;

                // nitems_ is even.
                if (sum > mid)
                    return i;
                // looking for next non null value
                unsigned i_memo = i;
                do
                    ++i;
                while (h_[i] == 0);
                if ((i_memo + i) % 2)
                    plus_half = true;
                return (i_memo + i) / 2;
            }


            bool is_valid() const
            {
                return nitems_ != 0;
            }

            protected:

            std::vector<unsigned> h_;
            unsigned nitems_;
        };

        template <typename T>
            struct imed< irange<T> > : public mln::accu::internal::base< unsigned, imed< irange<T> > >
            {
                typedef irange<T> argument;

                imed() {}

                void init()
                {
                    h_.resize(mln_max(T) + 1, 0);
                    nitems_ = 0;
                }

                void take(const argument& v)
                {
                    assert(v.is_degenerated());
                    h_[v.lower] += 1;
                    nitems_ += 1;
                }

                void take(const imed& other) {}

                unsigned to_result() const
                {
                    bool forget_it_;
                    return compute_now(forget_it_);
                }

                unsigned compute_now(bool& plus_half) const
                {
                    plus_half = false;
                    unsigned i = 0, n = mln_max(T) + 1;
                    unsigned sum = h_[0], mid = (nitems_ + 1) / 2;
                    while (i < n && sum < mid)
                    {
                        ++i;
                        sum += h_[i];
                    }
                    if (nitems_ % 2) // nitems_ is odd.
                        return i;

                    // nitems_ is even.
                    if (sum > mid)
                        return i;
                    // looking for next non null value
                    unsigned i_memo = i;
                    do
                        ++i;
                    while (h_[i] == 0);
                    if ((i_memo + i) % 2)
                        plus_half = true;
                    return (i_memo + i) / 2;
                }


                bool is_valid() const
                {
                    return nitems_ != 0;
                }

                protected:

                std::vector<unsigned> h_;
                unsigned nitems_;
            };


        template <typename T>
            std::ostream& operator<<(std::ostream& ostr, const imed<T>& m)
            {
                bool half;
                unsigned v = m.compute_now(half);
                if (half)
                    if (v == 0)
                        return ostr << "1/2";
                    else
                        return ostr << v << " 1/2";
                else
                    return ostr << v;
            }
    } // accu
} // mln
#endif // !ACCU_HH
