#ifndef CANON_HH_
# define CANON_HH_

#include <mln/core/site_set/p_set.hh>
#include "ufind_data.hh"
#include "max_tree.hh"

namespace mln {
    namespace K2 {
        struct is_primary_t : Function_v2b< is_primary_t > {
            typedef bool result;

            template <typename G, typename C>
                inline
                bool operator()(const point<G,C>& p) const {
                    for (int i = 0; i < G::dim; ++i)
                        if (p[i] % 4 != 0)
                            return false;
                    // mln_invariant(face_dim(p) == G::dim);
                    return true;
                }

            template <typename It>
                inline
                bool operator()(const Site_Iterator<It>& it) const {
                    typedef const mln_site(It)& P;
                    return this->operator()( exact(it).to_site() );
                }
        } is_primary;

        ufind_data* replace_representative(ufind_data* p, ufind_data* newrepr) {
            if (p->par == p || p->par->value != p->value) {
                if (is_primary(newrepr->pos) && !is_primary(p->pos)) {
                    if (p->par == p) {
                        newrepr->par = newrepr;
                    }
                    else {
                        newrepr->par = p->par;
                    }

                    p->par = newrepr;

                    return newrepr;
                }

                return p;
            }
            else {
                ufind_data* newpar;

                if (K2::is_primary(p->pos)) {
                    newpar = replace_representative(p->par, p);
                }
                else {
                    newpar = replace_representative(p->par, newrepr);
                }

                if (newpar != p) {
                    p->par = newpar;
                    return p->par;
                }

                return p;
            }
        }

        void canon_first_pass(SubtreeWorker& sw, unsigned ncols) {
            for (ufind_data* row : sw.getChildsRows()) {
                for (unsigned i = 0; i < ncols; ++i) {
                    ufind_data* udata = row + i;
                    if (udata->par != udata && udata->value == udata->par->value) {
                        replace_representative(udata, udata);
                    }
                }
            }
        }

        void canon_second_pass(SubtreeWorker& sw, unsigned ncols) {
            for (ufind_data* row : sw.getChildsRows()) {
                for (unsigned i = 0; i < ncols; ++i) {
                    ufind_data* udata = row + i;
                    if (udata->par->value != udata->value && udata->par->par->value == udata->par->value) {
                        // assert(K2::is_primary(udata->pos));
                        udata->par = udata->par->par;
                    }
                }
            }
        }
    }
}

#endif
