#ifndef IMMERSE_HH
# define IMMERSE_HH

#include <mln/core/alias/point2d.hh>
#include <mln/border/resize.hh>

#include "iqueue_parallel.hh"
#include "accu.hh"


namespace mln
{
    namespace K2
    {
        template <typename T>
            struct value_x_2
            { };

        template <unsigned n>
            struct value_x_2< value::int_u<n> >
            {
                typedef value::int_u<n+1> ret;
            };

        // domains
        template <typename B>
            inline
            B domain (const B& b)
            {
                mlc_is_a (B, Box)::check ();
                mln_deduce (B, site, delta) one;
                one.set_all (1);
                return B (4 * b.pmin () - one, 4 * b.pmax () + one);
            }

        /*
         * computes the grid, its border's median,
         * and pushes its borders' pixels in the queue.
         */
        template <typename T>
            unsigned
            compute_grid_return_median_queue(
                    const image2d<T>&                                             f,
                    image2d<irange<typename value_x_2<T>::ret > >&          F,
                    mln::iqueue_fast<unsigned , typename value_x_2<T>::ret>& q
                    )
            {
                typedef image2d<T> I;
                typedef const T* ptr_i;
                typedef typename value_x_2<T>::ret v_x_2;
                typedef irange<v_x_2> r_x_2;
                typedef r_x_2* ptr_r_x_2;

                int offset_s = F.delta_index(dpoint2d(1, 0));
                int offset_n = F.delta_index(dpoint2d(-1, 0));
                int offset_s_f = f.delta_index(dpoint2d(1, 0));
                int offset_n_f = f.delta_index(dpoint2d(-1, 0));

                ptr_i p_in;
                ptr_r_x_2 p_in_F;
                const unsigned nrows = f.nrows(), ncols = f.ncols();
                accu::imed<T> m;
                m.init();
                /*
                 * calculate nodes
                 */
                ptr_r_x_2 base_offset = F.buffer();
                /*
                 * p_ul . . . p_ur
                 * .    . . . p_ur_dr_N
                 * .    . . . p_ur_dr_M
                 * .    . . . p_ur_dr_S
                 * .    . . . p_dr
                 */
                /*
                 * Note: we could probably play with the fact that with:
                 * a b
                 * c d
                 * we will alway this pattern (if a > b > c > d)
                 * a a a ? d
                 * a a a ? d
                 * a a a ? d
                 * ? ? ? ? ?
                 * c c c ? d
                 * where ? is a openmin-span: ]min, max]
                 * The pattren is simpler (but still exact) if some values are equal.
                 */
                ptr_r_x_2 p_ul, p_ur, p_ur_dr_N, p_ur_dr_M, p_ur_dr_S, p_dr;
                ptr_r_x_2 ptr_start = &F.at_(0,0);
                ptr_i p_in_start = &f.at_(0, 0);
                p_ul = ptr_start;
                /* Set the value of:
                 *  V . . . .
                 *  . . . . .
                 *  . . . . .
                 *  . . . . .
                 *  . . . . .
                 */
                *p_ul = (2 * *p_in_start) ;
                m.take(*p_in_start); // should be optimized by the compiler
                /*
                 * for all lines, do
                 */
                for(int j = 0; j < nrows - 1; j++)
                {
                    /* A "cell" treated at the time is (I = Initial; M = Max; S = Span):
                     * I S M S I
                     * S S S S S
                     * M S M S M
                     * S S S S S
                     * I S M S I
                     * Notice that if we have to make a span with 4 elements,
                     * 2 are actually enouth! In:
                     * a . center
                     * . v .       =>  v = minospan(c, center) is enouth.
                     * c . d
                     */
                    p_in = p_in_start + (j + 1) * offset_s_f;
                    p_ur = p_ul + 4;
                    p_ur_dr_N = p_ur + offset_s;
                    p_ur_dr_M = p_ur_dr_N + offset_s;
                    p_ur_dr_S = p_ur_dr_M + offset_s;
                    p_dr = p_ur_dr_S + offset_s;

                    /*
                     * init first column
                     */
                    /* Set the value of:
                     *  | . . . .
                     *  . . . . .
                     *  . . . . .
                     *  . . . . .
                     *  V . . . .
                     */
                    *(p_dr - 4) = (2 * *p_in) ;
                    m.take(*p_in);
                    /* Set the value of:
                     *  | . . . .
                     *  V . . . .
                     *  V . . . .
                     *  V . . . .
                     *  | . . . .
                     */
                    v_x_2 v_ul = p_ul->lower;
                    v_x_2 v_dl = (p_dr - 4)->lower;
                    if(v_ul > v_dl)
                    {
                        *(p_ur_dr_M - 4) = v_ul;
                        *(p_ur_dr_N - 4) = v_ul; // recopy the max value
                        *(p_ur_dr_S - 4) = r_x_2(v_dl, v_ul);
                    }
                    else if(v_ul < v_dl)
                    {
                        *(p_ur_dr_M - 4) = v_dl;
                        *(p_ur_dr_S - 4) = v_dl; // recopy the max value
                        *(p_ur_dr_N - 4) = r_x_2(v_ul, v_dl);
                    }
                    else
                    {
                        *(p_ur_dr_M - 4) = v_ul;
                        *(p_ur_dr_N - 4) = v_ul;
                        *(p_ur_dr_S - 4) = v_ul;
                    }
                    p_in++;

                    /*
                     * queue
                     */
                    /*
                     * insertion in the queue.
                     * | . . . .
                     * | . . . .
                     * | . . . .
                     * | . . . .
                     * . . . . .
                     */
                    q.push_at_current_level(p_ul - 1 + offset_n - base_offset);
                    q.push_at_current_level(p_ul - 1 + offset_s - base_offset);
                    q.push_at_current_level(p_ul - 1 + 2 * offset_s - base_offset);
                    q.push_at_current_level(p_ul - 1 - base_offset);

                    /*
                     * for one line
                     */
                    for(int i = 0; i < ncols - 1; i++)
                    {
                        if(j == 0) // first line
                        {
                            /* Set the value of:
                             *  | . . . V
                             *  | . . . .
                             *  | . . . .
                             *  | . . . .
                             *  | . . . .
                             */
                            *p_ur = 2 * *(p_in + offset_n_f) ;
                            m.take(*(p_in + offset_n_f));
                            /* Set the value of:
                             *  | V V V V
                             *  | . . . .
                             *  | . . . .
                             *  | . . . .
                             *  | . . . .
                             */
                            v_x_2 v_ul = p_ul->lower;
                            v_x_2 v_ur = p_ur->lower;
                            if(v_ul < v_ur)
                            {
                                *(p_ul + 3) = v_ur;
                                *(p_ul + 2) = v_ur; // recopy the max value
                                *(p_ul + 1) = r_x_2(v_ul, v_ur);
                            }
                            else if(v_ur < v_ul)
                            {
                                *(p_ul + 1) = v_ul;
                                *(p_ul + 2) = v_ul; // recopy the max value
                                *(p_ul + 3) = r_x_2(v_ur, v_ul);
                            }
                            else // equality
                            {
                                *(p_ul + 1) = v_ur;
                                *(p_ul + 2) = v_ur;
                                *(p_ul + 3) = v_ur;
                            }

                            /*
                             * queue
                             */
                            q.push_at_current_level(p_ul + offset_n - base_offset);
                            q.push_at_current_level(p_ul + offset_n + 1 - base_offset);
                            q.push_at_current_level(p_ul + offset_n + 2 - base_offset);
                            q.push_at_current_level(p_ul + offset_n + 3 - base_offset);
                        }
                        /* Set the value of:
                         *  | | | | |
                         *  | . . . .
                         *  | . . . .
                         *  | . . . .
                         *  | . . . V
                         */
                        *p_dr = (2 * *p_in) ;
                        if(j == nrows - 2 || i == ncols - 2)
                            m.take((*p_in));
                        p_in ++;

                        /* Set the value of:
                         *  | | | | |
                         *  | . . . V
                         *  | . . . V
                         *  | . . . V
                         *  | . . . |
                         */
                        v_x_2 v_ur = p_ur->lower;
                        v_x_2 v_dr = p_dr->lower;
                        if(v_ur > v_dr)
                        {
                            *p_ur_dr_M = v_ur;
                            *p_ur_dr_N = v_ur; // recopy the max value
                            *p_ur_dr_S = r_x_2(v_dr, v_ur);
                        }
                        else if(v_ur < v_dr)
                        {
                            *p_ur_dr_M = v_dr;
                            *p_ur_dr_S = v_dr; // recopy the max value
                            *p_ur_dr_N = r_x_2(v_ur, v_dr);
                        }
                        else
                            *p_ur_dr_M = *p_ur_dr_N = *p_ur_dr_S = v_ur;

                        /* set the bottom line
                         *  | | | | |
                         *  | . . . |
                         *  | . . . |
                         *  | . . . |
                         *  | V V V |
                         */
                        v_x_2 v_dl = (p_dr - 4)->lower;
                        // v_x_2 v_dr = p_dr->lower; declared above
                        if(v_dl < v_dr)
                        {
                            *(p_dr - 1) = v_dr;
                            *(p_dr - 2) = v_dr; // recopy the max value
                            *(p_dr - 3) = r_x_2(v_dl, v_dr);
                        }
                        else if(v_dr < v_dl)
                        {
                            *(p_dr - 3) = v_dl;
                            *(p_dr - 2) = v_dl; // recopy the max value
                            *(p_dr - 1) = r_x_2(v_dr, v_dl);
                        }
                        else // equality
                        {
                            *(p_dr - 3) = v_dr;
                            *(p_dr - 2) = v_dr;
                            *(p_dr - 1) = v_dr;
                        }
                        /*
                         * set every missing values
                         * | | | | |
                         * | V V V |
                         * | V V V |
                         * | V V V |
                         * | | | | |
                         */
                        ptr_r_x_2 middle_ptr = p_ur_dr_M - 2;
                        v_x_2 middle_value = std::max((p_ur_dr_M - 4)->lower, (p_ur_dr_M)->lower);
                        *middle_ptr = middle_value;
                        for(int mi = -1; mi <= 1; mi++)
                        {
                            for(int mj = -1; mj <= 1; mj++)
                            {
                                if(mi || mj)
                                {
                                    v_x_2 corner_value = (middle_ptr + mi * 2 + mj * offset_n * 2)->lower;
                                    if(corner_value == middle_value)
                                        *(middle_ptr + mi + mj * offset_n) = middle_value;
                                    else
                                        *(middle_ptr + mi + mj * offset_n) = r_x_2(corner_value, middle_value);
                                }
                            }
                        }
                        p_ul    += 4;
                        p_ur    += 4;
                        p_ur_dr_N += 4;
                        p_ur_dr_M += 4;
                        p_ur_dr_S += 4;
                        p_dr    += 4;
                    }
                    /*
                     * insert right border's cells
                     */
                    if(j == 0)
                        q.push_at_current_level(p_ul + offset_n - base_offset);
                    q.push_at_current_level(p_ul + 1 + offset_n - base_offset);
                    q.push_at_current_level(p_ul + 1 - base_offset);
                    q.push_at_current_level(p_ul + 1 + offset_s - base_offset);
                    q.push_at_current_level(p_ul + 1 + 2 * offset_s - base_offset);
                    /*
                     * next line
                     */
                    p_ul = ptr_start + ((j + 1) * 4 * offset_s);
                }
                /*
                 * last line
                 */
                /*
                 * insert first border's cells
                 */
                q.push_at_current_level(p_ul - 1 + offset_n - base_offset);
                q.push_at_current_level(p_ul - 1 - base_offset);
                q.push_at_current_level(p_ul - 1 + offset_s - base_offset);
                q.push_at_current_level(p_ul + offset_s - base_offset);
                for(int i = 0; i < ncols - 1; i++)
                {
                    p_ul+=4;
                    q.push_at_current_level(p_ul + offset_s - base_offset);
                    q.push_at_current_level(p_ul - 1 + offset_s - base_offset);
                    q.push_at_current_level(p_ul - 2 + offset_s - base_offset);
                    q.push_at_current_level(p_ul - 3 + offset_s - base_offset);
                }
                /*
                 * insert last border's nodes
                 */
                q.push_at_current_level(p_ul + 1 + offset_s - base_offset);
                q.push_at_current_level(p_ul + 1 + offset_n - base_offset);
                q.push_at_current_level(p_ul + 1 - base_offset);
                bool plus_half;
                unsigned res = m.compute_now(plus_half);
                res = 2 * res + (plus_half ? 1 : 0);
                q.swap_levels(1, res);
                return res;
            }

        template <typename T>
            image2d< irange< typename value_x_2<T>::ret > >
            irange_immerse_opt_median_queue(
                    const image2d<T>&                                        f,
                    unsigned&                                                med,
                    mln::iqueue_fast< unsigned, typename value_x_2<T>::ret>& q,
                    bool                                                     change_thickness)
            {
                typedef image2d<T> I;

                image2d< irange< typename value_x_2<T>::ret > > F(K2::domain(f.domain()));
                if (change_thickness)
                    border::resize(F, 1);
                q.init(1, true);
                med = compute_grid_return_median_queue(f, F, q);
                return F;
            }
    } // end of namespace mln::K2
} // end of namespace mln

#endif // IMMERSE_HH
